# -*- mode:Python; tab-width: 4 -*-
"""Extract or transform links in HTML documents.

Link_extractor is a class which processes an HTML document and
stores all the external links. Link_transfromer, it's child, changes
external links to other options you provide with a dictionary.

Since processing is done through HTMLParser, bad HTML can break it.
The optional mx.Tidy module can be imported and used to clean_html
before parsing it, thus eliminating the danger of bad HTML.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""

__author__ = "Grzegorz Adam Hankiewicz <gradha@efaber.net>"
__date__ = "May 2003"
__version__ = "$Revision: 1.35 $"
__credits__ = ""

from HTMLParser import HTMLParser, HTMLParseError
import urlparse
import urllib
import re
import os

language_switch = re.compile(r"/language-prefs\?language_preference=(\w+)(.*)")
watched_tags = {"img": 'src', "a": 'href', "link": 'href',
    "td": 'background', "table": 'background', "tr": 'background',
    "body": 'background', "input": 'src', "form": 'action',
    "embed": 'src'}


def _tidy_error_message():
    print "*** mx.Tidy not installed, so I cannot parse invalid HTML."
    print "*** You can get mx.Tidy from http://www.egenix.com/files/python/mxTidy.html"
    raise Exception("mx.Tidy not available, but you wanted to clean HTML")

try:
    import mx.Tidy
except ImportError:
    pass


def split_all_path(path):
    """Splits 'path/to/directory/file' into ['path', 'to, 'directory', 'file'].

    Splits the given path and returns a list with the individual
    elements.
    """
    import os
    elements = []
    while path:
        head, tail = os.path.split(path)
        elements.insert(0, tail)
        if path == head:
            break
        path = head
    return filter(None, elements)


def split_url(url):
    """Cleans and splits an URL into URL and special fragment.

    URLs can have a fragment, which is the text after the hash in
    the URL to define a local link in the same page. This function
    first strips whitespace from the url, the splits it up into
    elements and reconstructs them without the fragment, which is
    returned as the second element of the final tuple.

    Note that it's not the usual fragment: urlparse.urlparse returns
    the fragment without the hash, this function returns the hash
    embedded in the fragment. This means that you can join the
    tuple returned by this function as a normal string concatenation
    operation and you will obtain the original URL cleaned up.

    If there is no tail, the empty string will be returned, which
    means that `"%s%s" % (url, tail)' will always return the
    expected value.
    """
    url = url.strip()
    t1, t2, t3, t4, t5, t6 = urlparse.urlparse(url)
    url, tail = urlparse.urlunparse((t1, t2, t3, t4, t5, "")), t6
    url = urllib.unquote(url)
    if tail:
        tail = "#%s" % tail
    return (url, tail)


def compose_attributes(pair_list):
    """Builds a string representing the pair_list HTML attributes.

    Transforms the HTML attribute pair list into a string which
    can be inserted in a tag.
    """
    final = [""]
    for key, value in pair_list:
        if value == None:
            final.append(key)
        else:
            final.append('%s="%s"' % (key, value))
    return " ".join(final)


def extract_pair_list_value(key, pair_list):
    """Get value from a (key, value) pair list specifying key.

    Treats a given pair list in form (key, value) as a
    dictionary. You pass the key, and the function returns the
    associated value. If key is not found, None is returned.
 
    Keys and values are strings.  The keys are compared in lower
    case, and values are returned after they have been stripped
    from leading and trailing whitespace.
    """
    searched_key = key.lower()
    for key, value in pair_list:
        if key.lower() == searched_key:
            return value.strip()
            
    return None


def relative_path(path_from, path_to, fragment = "", force_local_link = 0):
    """Returns a string with the relative path from path_from to path_to.

    The destination path has to be a real local file (either path_to
    or path_to.tmp), otherwise an IOError will be raised. If the two
    paths are on different drive the returned path will be invalid
    (ie: C:\... -> D:\...).

    The optional fragment parameter should be an empty string or
    one in the format '#whatever' which selects a local link in
    the final destination.

    The optional force_local_link boolean parameter is a special
    case required by the Link_transformer. If it is true, path_to
    won't be checked for validity, and IOError exceptions won't be
    raised if the file doesn't exist. In this case, .tmp won't be
    appended to the local path.
    """
    import os
    if not force_local_link:
        try:
            file = open(path_to, "rb")
        except IOError:
            file = open("%s.tmp" % path_to, "rb")
        file.close()
    path_from = split_all_path(os.path.realpath(os.path.dirname(path_from)))
    path_to = split_all_path(os.path.realpath(path_to))

    while path_from and path_to and path_from[0] == path_to[0]:
        path_from.pop(0)
        path_to.pop(0)

    final_path = [".."] * len(path_from) + path_to
    return "%s%s" % (os.path.join(*final_path), fragment)
    

class Link_extractor(HTMLParser):
    """Extract links from HTML code.
 
    Processes HTML code and stores all the external links, which
    you can get after the processing has finished.
    """
    def __init__(self, data, clean_html = 0):
        """Create the object parsing HTML immediately.

        After the object has been initialised, the object processes
        data, which is presumed to be a string containing the whole
        HTML code of a web page.

        If clean_html is not false and an HTMLParseException is
        raised during the initial parsing, the data will be feeded
        to mx.Tidy to clean it up and then the parsing will be
        retried. Note that mx.Tidy might modify your HTML code in
        unexpected ways, like not preserving whitespace you were
        counting on for HTML to be rendered visually as you expected.
  
        Once the object is returned you can extract its links or
        the processed HTML with the get_links and get_data methods.
        """
        HTMLParser.__init__(self)
        # tags which are watched for attributes with external references
        self._data = []
        self._links = {}
        try:
            self.feed(data)
            self.close()
        except HTMLParseError:
            self.reset()
            self.close()
            if clean_html:
                try:
                    f = mx.Tidy.tidy
                except NameError:
                    _tidy_error_message()
                nerrors, nwarnings, data, errordata = f(data)
                self._data = []
                self._links = {}
                self.feed(data)
                self.close()
            else:
                raise

    def handle_link(self, tag, attributes):
        """Handles the link extraction of an HTML tag.

        Given a tag and its attributes, tries to find any external
        reference and store it for posterior use.
        """
        try:
            select = watched_tags[tag.lower()]
            link = extract_pair_list_value(select, attributes)
        except KeyError:
            link = None

        if link and link[0] != '#':
            self._links[link] = 1

    def handle_starttag(self, tag, attributes):
        """Processes start tags.

        Appends the starting tag to the HTML buffer after extracting
        any external links from the tag's attributes.
        """
        attributes = self.transform_attributes(tag, attributes)
        self._data.append("<%s%s>" % (tag.lower(),
            compose_attributes(attributes)))
        self.handle_link(tag, attributes)

    def handle_endtag(self, tag):
        self._data.append("</%s>" % tag.lower())

    def handle_startendtag(self, tag, attributes):
        """Appends tags without ending to the buffer after processing them."""
        attributes = self.transform_attributes(tag, attributes)
        self._data.append("<%s%s />" % (tag.lower(),
            compose_attributes(attributes)))
        self.handle_link(tag, attributes)

    def transform_attributes(self, dummy, attributes):
        """Returns the attributes.

        This method is meant to be overridden in subclasses,
        it currently does nothing. It is called before any tag
        is appended to the HTML buffer, so you could use this to
        transform the attributes into something else.
        """
        return attributes

    def handle_data(self, data):
        self._data.append(data)

    def handle_charref(self, name):
        self._data.append("&#%s;" % name)

    def handle_entityref(self, name):
        self._data.append("&%s;" % name)

    def handle_comment(self, data):
        self._data.append("<!--%s-->" % data)

    def handle_decl(self, decl):
        self._data.append("<!%s>" % decl)

    def get_data(self):
        """Returns the processed HTML.

        Note that if the clean_html flag was true in the constructor
        call, the HTML could be different from the original data.
        """
        return "".join(self._data)

    def get_links(self):
        """Returns all the found external links."""
        return self._links.keys()

    def error(self, message):
        """Does nothing, other than making pychecker happy."""
        HTMLParser.error(self, message)

class Link_transformer(Link_extractor):
    """Transforms all the links in a given HTML document.

    Inherits from Link_extractor, working pretty much the same. The
    only difference is that this time you pass the constructors a
    few more parameters and links will be modified.
    """
    def __init__(self, local_file, url, substitutions, clean_html = 0,
                 forced_local_links = 0, relative_local_links = 1):
        """Create the object parsing HTML immediately.

        After the object has been initialised, the object processes
        an HTML file. If clean_html is not false, the data will be
        first feeded to mx.Tidy to clean it up. This avoids raising
        parser exceptions with bad HTML code.
  
        local_file is the path to the local file system where the
        HTML file is stored. url is the original URL of the very
        same file before it was saved to disk.
  
        substitutions is a dictionary. Its keys are the absolute
        URLs found in the opened file, and they should map to the
        absolute location of the local file, if any exists. Example:

        {"http://www.efaber.net/index.html": "out/test/index_c.html",
         "http://www.google.com/": "out/test/fake_google.html"}

        Don't worry about HTML code which uses relative links, that's
        why you give the url parameter in the constructor. The
        code first converts all relative links to absolute with
        the url parameter and *then* the link is searched in the
        substitution dictionery.  If a link is not found in the
        dictionary, it will be left as an external absolute link.

        forced_local_links is a boolean parameter. Usually, local
        links have to exist or the transformer will fail. Using
        here true, however, forces the transformation of absolute
        URLs to local non existent files. In these cases make sure
        substitutions has only the URLs you want to change.

        relative_local_links is a boolean parameter. If it is true,
        a relative path in the form ../../whatever will be created.
        If false, the final link passed in the substitutions
        directory won't be transformed.

        Once the object is returned you can extract the modified
        processed code with the get_data method.
        """
        self._change = substitutions
        self._url = url
        self._local_file = local_file
        self._forced_local_links = forced_local_links
        self._relative_local_links = relative_local_links
  
        file = open(local_file, "rb")
        data = file.read()
        file.close()
        
        Link_extractor.__init__(self, data, clean_html)

    def transform_attributes(self, tag, attributes):
        """Transforms any URLs in the attributes and returns them.

        Handles the transformation of the urls. First, the
        extracted link is converted to absolute and searched in
        the substitution dictionary. If found, a relative link is
        build from the file containing the HTML code to the file
        specified by the dictionary.

        As a special case, the url is searched for a language switch
        without forced redirectory part. If these conditions are
        met, a new relative link is tried to find the same name of
        the current file with the new langcode.
        """
        try:
            select = watched_tags[tag.lower()]
        except KeyError:
            return attributes
  
        # let's build a new attribute list
        new_attributes = []
        for key, value in attributes:
            if key.lower() == select:
                url, fragment = split_url(value)
                url = urlparse.urljoin(self._url, url)
                value = "%s%s" % (url, fragment)
                lang_code, redir = find_language_switch(value)
                if lang_code and not redir:
                    target = os.path.join(os.path.dirname(self._local_file),
                        os.path.basename(url_to_local_path(self._url)))
                    target = rename_local_path(target, lang_code)
                    if self._relative_local_links:
                        try:
                            value = relative_path(self._local_file, target, fragment)
                        except IOError:
                            pass
                    else:
                        value = "%s%s" % (target, fragment)
                else:
                    try:
                        path_to = self._change[url]
                        if self._relative_local_links:
                            value = relative_path(self._local_file,
                                path_to, fragment, self._forced_local_links)
                        else:
                            value = "%s%s" % (path_to, fragment)
                    except KeyError:
                        pass
  
            new_attributes.append((key, value))
              
        return new_attributes


def find_language_switch(url):
    """Returns lang_code, redirect if url switches language.

    If the URL is used to switch languages, returns a tuple with the
    new language code. If the URL features a redirect, it will be as
    the second element of the returned tuple.  If any of this data is
    not available, a tuple with two empty strings will be returned.
    """
    match = language_switch.search(url)
    if match:
        return match.group(1), match.group(2)
    return "", ""

# TODO: This is an ugly hack preventing reentrancy. Should be
# replaced with something more usable, like a passed parameter.
extension_hook = {}
html_extension = ".html"


def rename_local_path(filename, intra_extension = ""):
    """Transforms filename(.ext) into filename(.intra_extension).ext

    Given a local filename, this function inserts intra_extension
    between the head of the filename and it's extension, ONLY if
    the current file extension contains the string 'htm' or has no
    extension at all. Otherwise the same string is returned.
    """
    base, ext = os.path.splitext(filename)
    if not ext or ext.find(html_extension) >= 0:
        if intra_extension:
            try:
                filename = "%s%s%s" % (base, extension_hook[intra_extension],
                    html_extension)
            except KeyError:
                extension_hook[intra_extension] = ".%s" % intra_extension
                filename = "%s.%s%s" % (base, intra_extension, html_extension)
        else:
            filename = "%s%s" % (base, html_extension)
    return filename


def url_to_local_path(url, local_prefix = "", local_suffix = "index"):
    """Converts the remote path of the url to a local path.

    Given a complete URL, this function extracts the remote path
    element of the URL and returns it as a local relative path,
    optionally prefixing it with the directory contained in the
    string parameter local_prefix. If the resulting path doesn't
    contain a filename, local_suffix is appended to the result,
    which defaults to the string 'index'. Examples:

    url_to_local_path("http://www.efaber.net") -> 'index'
    url_to_local_path("http://www.efaber.net/contactos") -> 'contactos'
    url_to_local_path("http://www.efaber.net/menu/ayuda") -> 'menu/ayuda'
    url_to_local_path("http://www.efaber.net/menu") -> 'menu'
    url_to_local_path("http://www.efaber.net/menu/") -> 'menu/index'
    url_to_local_path("http://www.efaber.net/menu/", "output_test")
        -> 'menu/index'
    url_to_local_path("http://www.efaber.net/menu/", "output_test/")
        -> 'output_test/menu/index'
    url_to_local_path("http://www.efaber.net/menu/", "output_test/file.html")
        -> 'output_test/menu/index'
    url_to_local_path("http://www.efaber.net/menu/", "output_test/file.html",
        "indice") -> 'output_test/menu/indice'
    url_to_local_path("http://www.efaber.net/menu", "output_test/file.html",
        "indice") -> 'output_test/menu'

    Note that URL's parameters and query will be appended without
    separator.  Why? Well, browsers don't seem to agree on how to
    process HTML source code, so avoiding the parameter altogether
    gets rid of the problem. It also avoids weird characters in the
    local filename. There's a small chance of collision with a real
    URL being named like the localized version, but that should be
    pretty low in real life.
    """
    full = urlparse.urlparse(url)
    remote_path, parameters, query = full[2], full[3], full[4]
    remote_path = "".join([remote_path, parameters, query])
    local_path = os.path.join(os.path.dirname(local_prefix), remote_path[1:])
    if not os.path.basename(local_path):
        local_path = os.path.join(local_path, local_suffix)

    return local_path

